

class Otsu{
 public:
  int HIST_CLASS;
  int *myhist;
  
  
  
  Otsu(int val){
    HIST_CLASS = val;
    myhist = new int[HIST_CLASS];
  }
  
  virtual ~Otsu(){
    delete [] myhist;
  }
  
  
    
  int find_otsu_threshold(Image *in){
    histogram(in,HIST_CLASS,myhist);
    return find_hist_threshold(HIST_CLASS,myhist);
  }
  
  
      
 private:
  void histogram(Image *in, int hsize, int *hist){
    int i,x,y;
    for(i=0;i<hsize;i++)hist[i]=0;
    
    for(y=0;y<in->sy;y++)
      for(x=0;x<in->sx;x++){
	int val = ((int)(in->img[y][x]));
	if(val>HIST_CLASS-1)val=HIST_CLASS-1;
	if(val<0)val=0;
	
	hist[val]++;
      }
  }
    
      
  
  bool find_hist_range(int hsize, int *hist, int *class_min, int *class_max)
  {
    int i;
    
    for (i = 0; i < hsize; i++) {
      if (hist[i] != 0)
	break;
    }
    if (i >= hsize) {
      return false;  
    }
    *class_min = i;
    
    for (i=hsize-1; i>=(*class_min); i--) {
      if (hist[i] != 0)
	break;
    }
    *class_max = i;
    
    return true;
  }
  
  int find_hist_threshold(int hsize, int *hist){
    int num_class, class_min, class_max;
    long long count_lower, data_lower, count_upper, data_upper, thresh;
    double mean_lower, mean_upper,
      *distrib, max_dist;
    int   i, j;
    
    if (! find_hist_range(hsize, hist, &class_min, &class_max)) {
      return (hsize + 1) / 2;  
    }
    num_class = class_max - class_min + 1;
    if (num_class == 1) {
      return class_max;  
      }
      if (num_class == 2) {
	return class_max;
      }
      
      distrib = new double[(num_class - 1)];
      
      count_lower = data_lower = count_upper = data_upper = 0;
      
      for (i=class_min; i<=class_max; i++) {
	count_upper  += hist[i];
	data_upper   += hist[i] * i;
      }
      
      for (i=class_min,j=0; i<class_max; i++,j++) { 
	
	count_lower += hist[i];
	data_lower  += hist[i] * i;
	count_upper -= hist[i];
	data_upper  -= hist[i] * i;
	
	mean_lower = (count_lower != 0) ?
	  (double)data_lower / count_lower : 0.0;
	mean_upper = (count_upper != 0) ?
	  (double)data_upper / count_upper : 0.0;
	distrib[j] = count_lower * count_upper *
	  (mean_upper - mean_lower) * (mean_upper - mean_lower);
      }
      
      thresh = class_min + 1;
      max_dist = distrib[0];
      for (i=1; i<num_class-1; i++) {
	if (distrib[i] > max_dist) {
	  thresh = i + class_min + 1;
	  max_dist = distrib[i];
	}
      }
      
      delete [] distrib;
      return thresh;
   }
  

  Otsu(const Otsu& rhs);
  const Otsu &operator=(const Otsu& rhs);
 
 
};

#undef PUT
